
#define BITS 8	//    
#define MAX  ((1<<BITS)-1)
#define EPS 1e-8
#define PI 3.1415926
#define POW3(x) x*x*x

#define RX  95.047     // Observer = 2, Illuminant = D65(Daylight)
#define RY 100.000
#define RZ 108.883

#include<math.h>
#include<stdio.h>
#include<assert.h>

struct Lab{double l;double a;double b;};
struct Luv{double l;double u;double v;};
struct LCH{double l;double c;double h;};
struct XYZ{double x;double y;double z;};
struct Yxy{double Y;double x;double y;};
struct HSV{double h;double s;double v;};
struct HSL{double h;double s;double l;};
struct CMYK{int c;int m;int y;int k;};
struct RGB{int r;int g;int b;};

const class ColorSpace{
private:
	int Hue_to_RGB(double v1,double v2,double h);
	double RGB_to_Hue(double r,double g,double b,double del,double max);
protected:
	bool isZero(double x) {return fabs(x)<EPS;}
	template<typename T>T max(T x,T y) {return x<y?y:x;}
	template<typename T>T min(T x,T y) {return x<y?x:y;}
	template<typename T>T max(T x,T y,T z) {return max(x,max(y,z));}
	template<typename T>T min(T x,T y,T z) {return min(x,min(y,z));}
	void RotBound(double&);
	void CutBound(int&);
	void CutRGB(RGB&);
public:
	void XYZ_to_Yxy(XYZ,Yxy&);
	void Yxy_to_XYZ(Yxy,XYZ&);
	void RGB_to_XYZ(RGB,XYZ&);
	void XYZ_to_RGB(XYZ,RGB&);
	void RGB_to_HSV(RGB,HSV&);
	void HSV_to_RGB(HSV,RGB&);
	void RGB_to_HSL(RGB,HSL&);
	void HSL_to_RGB(HSL,RGB&);
	void RGB_to_Lab_Hunter(RGB,Lab&);
	void Lab_to_RGB_Hunter(Lab,RGB&);
	void RGB_to_Lab_Cie(RGB,Lab&);
	void Lab_to_RGB_Cie(Lab,RGB&);
	void RGB_to_CYMK(RGB,CMYK&);
	void CYMK_to_RGB(CMYK,RGB&);
	void Lab_to_LCH_Cie(Lab,LCH&);
	void LCH_to_Lab_Cie(LCH,Lab&);
	void Luv_to_XYZ_Cie(Luv,XYZ&);
	void XYZ_to_Luv_Cie(XYZ,Luv&);
};


void ColorSpace::RotBound(double &x)
{
	if(x<0) x+=1;
	if(x>1) x-=1;
}


void ColorSpace::CutBound(int &x)
{
	if(x<0  ) x=0;
	if(x>MAX) x=MAX;
}

void ColorSpace::CutRGB(RGB &rgb)
{
	CutBound(rgb.r);
	CutBound(rgb.g);
	CutBound(rgb.b);
}


int ColorSpace::Hue_to_RGB(double v1,double v2,double h)
{
	double v;
	RotBound(h);
	if     (h<1.0/6) v=v1+(v2-v1)*h*6;
	else if(h<1.0/2) v=v2;
	else if(h<2.0/3) v=v1+(v2-v1)*(2.0/3-h)*6;
	else v=v1;
	return int(v*MAX);
}

inline double ColorSpace::RGB_to_Hue(double r,double g,double b,double del,double max)
{
	double h;
#if 0
	double dr=((max-r)/6+del/2)/del;
	double dg=((max-g)/6+del/2)/del;
	double db=((max-b)/6+del/2)/del;
	if      (r==max) h=db-dg;
	else if (g==max) h=1.0/3+dr-db;
	else if (b==max) h=2.0/3+dg-dr;
	else assert(0);
#else	//    
	del*=6;
	if      (r==max) h=      (g-b)/del;
	else if (g==max) h=1.0/3+(b-r)/del;
	else if (b==max) h=2.0/3+(r-g)/del;
	else assert(0);
#endif
	RotBound(h);
	return h;
}



void ColorSpace::XYZ_to_Yxy(XYZ in,Yxy &out)
{
	double s=in.x+in.y+in.z;
	out.Y=in.y;
	out.x=in.x/s;
	out.y=in.y/s;
}


void ColorSpace::Yxy_to_XYZ(Yxy in,XYZ &out)
{
	double d=in.Y/in.y;
	out.x=in.x*d;
	out.y=in.Y;
	out.z=(1-in.x-in.y)*d;
}

void ColorSpace::RGB_to_XYZ(RGB in,XYZ &out)
{
#define RGB_XYZ_c(x) if(x>0.04045) x=pow((x+0.055)/1.055,2.4); else x/=12.92; x*=100
	double r=(double)in.r/MAX;
	double g=(double)in.g/MAX;
	double b=(double)in.b/MAX;
	RGB_XYZ_c(r);
	RGB_XYZ_c(g);
	RGB_XYZ_c(b);
	//    ,    
	out.x=r*0.4124+g*0.3576+b*0.1805;	
	out.y=r*0.2126+g*0.7152+b*0.0722;
	out.z=r*0.0193+g*0.1192+b*0.9505;
}

void ColorSpace::XYZ_to_RGB(XYZ in,RGB& out)
{
#define XYZ_RGB_c(x) if(x>0.0031308) x=1.055*pow(x,1/2.4)-0.055; else x*=12.92
	double x=in.x/100;   
	double y=in.y/100;   
	double z=in.z/100;   
	//    ,    
	double r=x* 3.2406+y*-1.5372+z*-0.4986;
	double g=x*-0.9689+y* 1.8758+z* 0.0415;
	double b=x* 0.0557+y*-0.2040+z* 1.0570;
	XYZ_RGB_c(r);
	XYZ_RGB_c(g);
	XYZ_RGB_c(b);
	out.r=int(r*MAX);
	out.g=int(g*MAX);
	out.b=int(b*MAX);
	CutRGB(out);
}

void ColorSpace::RGB_to_HSV(RGB in,HSV &out)
{
	double r=(double)in.r/MAX;
	double g=(double)in.g/MAX;
	double b=(double)in.b/MAX;
	double min=ColorSpace::min(r,g,b);
	double max=ColorSpace::max(r,g,b);
	double del=max-min;
	out.v=max;
	if(isZero(del)) out.h=out.s=0;
	else {
		out.s=del/max;
		out.h=RGB_to_Hue(r,g,b,del,max);
	}
}

void ColorSpace::HSV_to_RGB(HSV in,RGB &out)
{
	if(isZero(in.s))                       //HSV from 0 to 1
	{
		out.r=out.g=out.b=int(in.v*MAX);
	}else{
		double r,g,b;
		double h=in.h*6;
		if(h==6) h=0;
		int i=(int)h;
		double v1=in.v*(1-in.s);
		double v2=in.v*(1-in.s*(h-i));
		double v3=in.v*(1-in.s*(1-(h-i)));
		switch(i) {
		case 0: r=in.v; g=v3  ; b=v1  ; break;
		case 1: r=v2  ; g=in.v; b=v1  ; break;
		case 2: r=v1  ; g=in.v; b=v3  ; break;
		case 3: r=v1  ; g=v2  ; b=in.v; break;
		case 4: r=v3  ; g=v1  ; b=in.v; break;
		default:r=in.v; g=v1  ; b=v2;   break;
		}
		out.r=int(r*MAX); 
		out.g=int(g*MAX);
		out.b=int(b*MAX);
	}
	CutRGB(out);
}

void ColorSpace::RGB_to_HSL(RGB in,HSL& out)
{
	double r=(double)in.r/MAX;
	double g=(double)in.g/MAX;
	double b=(double)in.b/MAX;
	double min=ColorSpace::min(r,g,b);
	double max=ColorSpace::max(r,g,b);
	double del=max-min;
	out.l=(max+min)/2;
	if(isZero(del)) out.h=out.s=0;
	else {                                   //Chromatic data...
		if(out.l<0.5) out.s=del/(max+min); else out.s=del/(2-max-min); //     
		out.h=RGB_to_Hue(r,g,b,del,max);
	}
}

void ColorSpace::HSL_to_RGB(HSL in,RGB& out)
{
	if(isZero(in.s))                       //HSL from 0 to 1
	{
		out.r=out.g=out.b=int(in.l*MAX);
	}else{
		double v1,v2;
		if(in.l<0.5) v2=in.l*(1+in.s); else v2=in.l+in.s-(in.s*in.l); //     
		v1=2*in.l-v2;
		out.r=Hue_to_RGB(v1,v2,in.h+1.0/3);
		out.g=Hue_to_RGB(v1,v2,in.h      );
		out.b=Hue_to_RGB(v1,v2,in.h-1.0/3);
	}
	CutRGB(out);
}

void ColorSpace::RGB_to_Lab_Hunter(RGB in,Lab &out)
{
	XYZ xyz;
	RGB_to_XYZ(in,xyz);
	double qy=sqrt(xyz.y);
	out.l=10*qy;
	out.a=17.5*(1.02*xyz.x-xyz.y)/qy;
	out.b=7*(xyz.y-0.847*xyz.z)/qy;
}

void ColorSpace::Lab_to_RGB_Hunter(Lab in,RGB &out)
{
	XYZ xyz;
	double y=in.l/10;
	double x=in.a/17.5*in.l/10;
	double z=in.b/ 7  *in.l/10;
	xyz.y=y*y;
	xyz.x=(x+xyz.y)/1.02;
	xyz.z=(xyz.y-z)/0.847;
	XYZ_to_RGB(xyz,out);
}

void ColorSpace::RGB_to_Lab_Cie(RGB in,Lab &out)
{
#define RGB_LAB_c(x) if(x>0.008856) x=pow(x,1.0/3);else x=x*7.787+16.0/116
	XYZ xyz;
	RGB_to_XYZ(in,xyz);
	double x=xyz.x/RX;  //        
	double y=xyz.y/RY;         
	double z=xyz.z/RZ;
	RGB_LAB_c(x);
	RGB_LAB_c(y);
	RGB_LAB_c(z);
	out.l=y*116-16;
	out.a=(x-y)*500;
	out.b=(y-z)*200;
}

void ColorSpace::Lab_to_RGB_Cie(Lab in,RGB &out)
{
#define LAB_RGB_c(x) if(POW3(x)>0.008856) x=POW3(x); else x=(x-16.0/116)/7.787
	XYZ xyz;
	xyz.y=(in.l+16)/116;
	xyz.x=(in.a/500)+xyz.y;
	xyz.z=xyz.y-in.b/200;
	LAB_RGB_c(xyz.x);
	LAB_RGB_c(xyz.y);
	LAB_RGB_c(xyz.z);
	xyz.x*=RX;    
	xyz.y*=RY;
	xyz.z*=RZ;
	XYZ_to_RGB(xyz,out);
}



void ColorSpace::Lab_to_LCH_Cie(Lab in,LCH &out)
{
	double h=atan2(in.b,in.a);
	if(h>0) h=(h/PI)*180; else h=360-(fabs(h)/PI)*180;
	out.l=in.l;
	out.c=sqrt(in.a*in.a+in.b*in.b);
	out.h=h;
}

void ColorSpace::LCH_to_Lab_Cie(LCH in,Lab &out)
{
	out.l=in.l;
	out.a=cos(in.h/180*PI)*in.c;
	out.b=sin(in.h/180*PI)*in.c;
}

void ColorSpace::Luv_to_XYZ_Cie(Luv in,XYZ &out)
{
	double y=(in.l+16)/116;
	if(POW3(y)>0.008856) y=POW3(y); else y=(y-16.0/116)/ 7.787;
	double t=RX+RY*15+RZ*3;
	double ru=RX*4/t;
	double rv=RY*9/t;
	double u=in.u/(in.l*13)+ru;
	double v=in.v/(in.l*13)+rv;
	out.y=y*100;
	out.x=u*out.y*9/(4*v);
	out.z=(out.y*9/v-out.y*15-out.x)/3;
}

void ColorSpace::XYZ_to_Luv_Cie(XYZ in,Luv &out)
{
	double t=in.x+in.y*15+in.z*3;
	double u=in.x*4/t;
	double v=in.y*9/t;
	double y=in.y/100;
	if(y>0.008856) y=pow(y,1.0/3); else y=y*7.787+16.0/116;
	t=RX+RY*15+RZ*3;
	double ru=RX*4/t;
	double rv=RY*9/t;
	out.l=y*116-16;
	out.u=out.l*13*(u-ru);
	out.v=out.l*13*(v-rv);
}

void ColorSpace::RGB_to_CYMK(RGB in,CMYK &out)
{
	out.c=MAX-in.r;
	out.m=MAX-in.g;
	out.y=MAX-in.b;
	out.k=min(out.c,out.m,out.y);   //    
	if(out.k>=MAX) out.c=out.m=out.y=0;
	else {
		out.c=(out.c-out.k)*MAX/(MAX-out.k);
		out.m=(out.m-out.k)*MAX/(MAX-out.k);
		out.y=(out.y-out.k)*MAX/(MAX-out.k);
	}
}


void ColorSpace::CYMK_to_RGB(CMYK in,RGB &out)
{
	out.r=MAX-in.c*(MAX-in.k)/MAX-in.k;
	out.g=MAX-in.m*(MAX-in.k)/MAX-in.k;
	out.b=MAX-in.y*(MAX-in.k)/MAX-in.k;
	CutRGB(out);
}


void main()
{
	RGB rgb;
	XYZ xyz; 
	Yxy yxy;
	HSV hsv;
	HSL hsl;
	Luv luv;
	Lab lab;
	LCH lch;
	CMYK cmyk;
	ColorSpace sp;

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_XYZ(rgb,xyz);
	sp.XYZ_to_RGB(xyz,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>1) _asm int 3

	xyz.x=123; xyz.y=45; xyz.z=67;
	sp.XYZ_to_Yxy(xyz,yxy);
	sp.Yxy_to_XYZ(yxy,xyz);
	if(fabs(xyz.x+xyz.y+xyz.z-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_HSV(rgb,hsv);
	sp.HSV_to_RGB(hsv,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_HSL(rgb,hsl);
	sp.HSL_to_RGB(hsl,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>2) _asm int 3

	xyz.x=123; xyz.y=45; xyz.z=67;
	sp.XYZ_to_Luv_Cie(xyz,luv);
	sp.Luv_to_XYZ_Cie(luv,xyz);
	if(fabs(xyz.x+xyz.y+xyz.z-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_Lab_Hunter(rgb,lab);
	sp.Lab_to_RGB_Hunter(lab,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_Lab_Cie(rgb,lab);
	sp.Lab_to_LCH_Cie(lab,lch);
	sp.LCH_to_Lab_Cie(lch,lab);
	sp.Lab_to_RGB_Cie(lab,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_CYMK(rgb,cmyk);
	sp.CYMK_to_RGB(cmyk,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>2) _asm int 3
}



